/*
Module : Sudoku.h
Purpose: Defines a collection of C++ classes which implement support for the classic 9*9 Sudoku puzzle
Created: PJN / 03-11-2013
History: PJN / 17-05-2016 1. Updated copyright details.
                          2. Updated the sample app to compile cleanly using /analyze
                          3. Updated the code to compile cleanly on VC 2015
         PJN / 30-12-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes
                          mean that the code will now only compile on VC 2017 or later.
         PJN / 04-04-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 08-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2013 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once


//////////////////// Includes /////////////////////////////////////////////////

#ifndef _ARRAY_
#pragma message("To avoid this message, please put array in your pre compiled header (usually stdafx.h)")
#include <array>
#endif //#ifndef _ARRAY_
#ifndef _INC_LIMITS
#pragma message("To avoid this message, please put limits.h in your pre compiled header (usually stdafx.h)")
#include <limits.h>
#endif
#include <cassert>


//////////////////// Classes //////////////////////////////////////////////////

class CSudokuGrid
{
public:
//Methods
  bool operator==(const CSudokuGrid& grid) noexcept
  {
    for (size_t i{0}; i<m_Grid.size(); i++)
    {
#pragma warning(suppress: 26446 26482)
      const auto& element1{m_Grid[i]};
      for (size_t j{0}; j<element1.size(); j++)
      {
#pragma warning(suppress: 26446 26482)
        const auto& element2{element1[j]};
#pragma warning(suppress: 26446 26482)
        if (element2 != grid.m_Grid[i][j])
          return false;
      }
    }
    return true;
  }

  bool operator!=(const CSudokuGrid& grid) noexcept
  {
    return !operator==(grid);
  }

  void CreateStandardSolvedGrid() noexcept
  {
#pragma warning(suppress: 26446)
    m_Grid[0][0] = 1;
#pragma warning(suppress: 26446)
    m_Grid[0][1] = 2;
#pragma warning(suppress: 26446)
    m_Grid[0][2] = 3;
#pragma warning(suppress: 26446)
    m_Grid[0][3] = 4;
#pragma warning(suppress: 26446)
    m_Grid[0][4] = 5;
#pragma warning(suppress: 26446)
    m_Grid[0][5] = 6;
#pragma warning(suppress: 26446)
    m_Grid[0][6] = 7;
#pragma warning(suppress: 26446)
    m_Grid[0][7] = 8;
#pragma warning(suppress: 26446)
    m_Grid[0][8] = 9;

#pragma warning(suppress: 26446)
    m_Grid[1][0] = 4;
#pragma warning(suppress: 26446)
    m_Grid[1][1] = 5;
#pragma warning(suppress: 26446)
    m_Grid[1][2] = 6;
#pragma warning(suppress: 26446)
    m_Grid[1][3] = 7;
#pragma warning(suppress: 26446)
    m_Grid[1][4] = 8;
#pragma warning(suppress: 26446)
    m_Grid[1][5] = 9;
#pragma warning(suppress: 26446)
    m_Grid[1][6] = 1;
#pragma warning(suppress: 26446)
    m_Grid[1][7] = 2;
#pragma warning(suppress: 26446)
    m_Grid[1][8] = 3;

#pragma warning(suppress: 26446)
    m_Grid[2][0] = 7;
#pragma warning(suppress: 26446)
    m_Grid[2][1] = 8;
#pragma warning(suppress: 26446)
    m_Grid[2][2] = 9;
#pragma warning(suppress: 26446)
    m_Grid[2][3] = 1;
#pragma warning(suppress: 26446)
    m_Grid[2][4] = 2;
#pragma warning(suppress: 26446)
    m_Grid[2][5] = 3;
#pragma warning(suppress: 26446)
    m_Grid[2][6] = 4;
#pragma warning(suppress: 26446)
    m_Grid[2][7] = 5;
#pragma warning(suppress: 26446)
    m_Grid[2][8] = 6;

#pragma warning(suppress: 26446)
    m_Grid[3][0] = 2;
#pragma warning(suppress: 26446)
    m_Grid[3][1] = 3;
#pragma warning(suppress: 26446)
    m_Grid[3][2] = 4;
#pragma warning(suppress: 26446)
    m_Grid[3][3] = 5;
#pragma warning(suppress: 26446)
    m_Grid[3][4] = 6;
#pragma warning(suppress: 26446)
    m_Grid[3][5] = 7;
#pragma warning(suppress: 26446)
    m_Grid[3][6] = 8;
#pragma warning(suppress: 26446)
    m_Grid[3][7] = 9;
#pragma warning(suppress: 26446)
    m_Grid[3][8] = 1;

#pragma warning(suppress: 26446)
    m_Grid[4][0] = 5;
#pragma warning(suppress: 26446)
    m_Grid[4][1] = 6;
#pragma warning(suppress: 26446)
    m_Grid[4][2] = 7;
#pragma warning(suppress: 26446)
    m_Grid[4][3] = 8;
#pragma warning(suppress: 26446)
    m_Grid[4][4] = 9;
#pragma warning(suppress: 26446)
    m_Grid[4][5] = 1;
#pragma warning(suppress: 26446)
    m_Grid[4][6] = 2;
#pragma warning(suppress: 26446)
    m_Grid[4][7] = 3;
#pragma warning(suppress: 26446)
    m_Grid[4][8] = 4;

#pragma warning(suppress: 26446)
    m_Grid[5][0] = 8;
#pragma warning(suppress: 26446)
    m_Grid[5][1] = 9;
#pragma warning(suppress: 26446)
    m_Grid[5][2] = 1;
#pragma warning(suppress: 26446)
    m_Grid[5][3] = 2;
#pragma warning(suppress: 26446)
    m_Grid[5][4] = 3;
#pragma warning(suppress: 26446)
    m_Grid[5][5] = 4;
#pragma warning(suppress: 26446)
    m_Grid[5][6] = 5;
#pragma warning(suppress: 26446)
    m_Grid[5][7] = 6;
#pragma warning(suppress: 26446)
    m_Grid[5][8] = 7;

#pragma warning(suppress: 26446)
    m_Grid[6][0] = 3;
#pragma warning(suppress: 26446)
    m_Grid[6][1] = 4;
#pragma warning(suppress: 26446)
    m_Grid[6][2] = 5;
#pragma warning(suppress: 26446)
    m_Grid[6][3] = 6;
#pragma warning(suppress: 26446)
    m_Grid[6][4] = 7;
#pragma warning(suppress: 26446)
    m_Grid[6][5] = 8;
#pragma warning(suppress: 26446)
    m_Grid[6][6] = 9;
#pragma warning(suppress: 26446)
    m_Grid[6][7] = 1;
#pragma warning(suppress: 26446)
    m_Grid[6][8] = 2;

#pragma warning(suppress: 26446)
    m_Grid[7][0] = 6;
#pragma warning(suppress: 26446)
    m_Grid[7][1] = 7;
#pragma warning(suppress: 26446)
    m_Grid[7][2] = 8;
#pragma warning(suppress: 26446)
    m_Grid[7][3] = 9;
#pragma warning(suppress: 26446)
    m_Grid[7][4] = 1;
#pragma warning(suppress: 26446)
    m_Grid[7][5] = 2;
#pragma warning(suppress: 26446)
    m_Grid[7][6] = 3;
#pragma warning(suppress: 26446)
    m_Grid[7][7] = 4;
#pragma warning(suppress: 26446)
    m_Grid[7][8] = 5;

#pragma warning(suppress: 26446)
    m_Grid[8][0] = 9;
#pragma warning(suppress: 26446)
    m_Grid[8][1] = 1;
#pragma warning(suppress: 26446)
    m_Grid[8][2] = 2;
#pragma warning(suppress: 26446)
    m_Grid[8][3] = 3;
#pragma warning(suppress: 26446)
    m_Grid[8][4] = 4;
#pragma warning(suppress: 26446)
    m_Grid[8][5] = 5;
#pragma warning(suppress: 26446)
    m_Grid[8][6] = 6;
#pragma warning(suppress: 26446)
    m_Grid[8][7] = 7;
#pragma warning(suppress: 26446)
    m_Grid[8][8] = 8;
  }

  bool IsLegalRow(int nRow, int nCol, unsigned char byValue) const noexcept
  {
    //Validate our parameters
    assert(byValue >= 0);
    assert(byValue <= 9);

    for (int i{0}; i<9; i++)
    {
#pragma warning(suppress: 26446 26482)
      if ((nCol != i) && (m_Grid[nRow][i] == byValue) && (byValue != 0))
        return false;
    }

    return true;
  }

  bool IsLegalColumn(int nRow, int nCol, unsigned char byValue) const noexcept
  {
    //Validate our parameters
    assert(byValue >= 0);
    assert(byValue <= 9);

    for (int i{0}; i<9; i++)
    {
#pragma warning(suppress: 26446 26482)
      if ((nRow != i) && (m_Grid[i][nCol] == byValue) && (byValue != 0))
        return false;
    }

    return true;
  }

  bool IsLegalBlock(int nRow, int nCol, unsigned char byValue) const noexcept
  {
    //Validate our parameters
    assert(byValue >= 0);
    assert(byValue <= 9);

    const int nStartingCol{(nCol / 3) * 3};
    const int nStartingRow{(nRow / 3) * 3};

    for (int i{0}; i<3; i++) //i is the row index
    {
      for (int j{0}; j<3; j++) //j is the column index
      {
        const int nRowToCheck{i + nStartingRow};
        const int nColToCheck{j + nStartingCol};

#pragma warning(suppress: 26446 26482)
        if ((nRow != nRowToCheck) && (nCol != nColToCheck) && (m_Grid[nRowToCheck][nColToCheck] == byValue) && (byValue != 0))
          return false;
      }
    }

    return true;
  }

  int IsCellEmpty(int nRow, int nCol) const noexcept
  {
    return GetCell(nRow, nCol) == 0;
  }

  void SetCell(int nRow, int nCol, unsigned char byValue) noexcept
  {
    //Validate our parameters
    assert(byValue >= 0);
    assert(byValue <= 9);

#pragma warning(suppress: 26446 26482)
    m_Grid[nRow][nCol] = byValue;
  }

  void SetCellEmpty(int nRow, int nCol) noexcept
  {
    SetCell(nRow, nCol, 0);
  }

  void EmptyAllCells() noexcept
  {
    //Empty each cell
    for (int i{0}; i<9; i++) //i is row index
    {
      for (int j{0}; j<9; j++) //j is column index
        SetCellEmpty(i, j);
    }
  }

  bool IsSolved() const noexcept
  {
    for (int i{0}; i<9; i++) //i is row index index
    {
      for (int j{0}; j<9; j++) //j is column index index
      {
        if (IsCellEmpty(i, j))
          return false;
      }
    }

    return true;
  }

  unsigned char GetCell(int nRow, int nCol) const noexcept
  {
#pragma warning(suppress: 26446 26482)
    return m_Grid[nRow][nCol];
  }

  int EmptyCells() const noexcept
  {
    //What will be the return value from this method
    int nEmptyCells{0};

    for (int i{0}; i<9; i++) //i is row index index
    {
      for (int j{0}; j<9; j++) //j is column index index
      {
        if (IsCellEmpty(i, j))
          ++nEmptyCells;
      }
    }

    return nEmptyCells;
  }

  int FilledCells() const noexcept
  {
    return 81 - EmptyCells();
  }

  int ChoicesForCell(int nRow, int nCol, unsigned char* pbyOnlyChoice) noexcept
  {
    //What will be the return value from this method
    int nChoices{0};

    unsigned char byNumber{0};
    if (IsCellEmpty(nRow, nCol))
    {
      for (unsigned char k{1}; (k<=9); k++)
      {
        if (IsLegalColumn(nRow, nCol, k) && IsLegalRow(nRow, nCol, k) && IsLegalBlock(nRow, nCol, k))
        {
          ++nChoices;
          byNumber = k;
        }
      }
    }

    //Fill in the output parameter if provided
    if ((nChoices == 1) && (pbyOnlyChoice != nullptr))
    {
      *pbyOnlyChoice = byNumber;
    }

    return nChoices;
  }

  bool FindEmptyCellWithLeastChoice(int& nRow, int& nCol, int& nChoices) noexcept
  {
    //What will be the return value from this method
    bool bFound{false};

    nChoices = INT_MAX;
    nRow = -1;
    nCol = -1;

    for (int i{0}; i<9; i++) //i is row index index
    {
      for (int j{0}; j<9; j++) //j is column index index
      {
        if (IsCellEmpty(i, j))
        {
          int nThisCellChoices{ChoicesForCell(i, j, nullptr)};
#pragma warning(suppress: 26489)
          if (nThisCellChoices < nChoices)
          {
            bFound = true;
            nChoices = nThisCellChoices;
            nRow = i;
            nCol = j;
          }
        }
      }
    }

    return bFound;
  }

  void SwapRows(int nRow1, int nRow2) noexcept //Swap the contents of one row with another
  {
    //Validate our parameters
    assert(nRow1 != nRow2);

    //Make a copy of the row first
#pragma warning(suppress: 26446 26482)
    std::swap(m_Grid[nRow1], m_Grid[nRow2]);
  }

  void SwapColumns(int nCol1, int nCol2) noexcept //Swap the contents of one column with another
  {
    //Validate our parameters
    assert(nCol1 != nCol2);

    //Make a copy of the column first
    std::array<unsigned char, 9> tempCol;
#pragma warning(suppress: 26446 26482)
    tempCol[0] = m_Grid[0][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[1] = m_Grid[1][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[2] = m_Grid[2][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[3] = m_Grid[3][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[4] = m_Grid[4][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[5] = m_Grid[5][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[6] = m_Grid[6][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[7] = m_Grid[7][nCol1];
#pragma warning(suppress: 26446 26482)
    tempCol[8] = m_Grid[8][nCol1];

#pragma warning(suppress: 26446 26482)
    m_Grid[0][nCol1] = m_Grid[0][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[1][nCol1] = m_Grid[1][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[2][nCol1] = m_Grid[2][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[3][nCol1] = m_Grid[3][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[4][nCol1] = m_Grid[4][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[5][nCol1] = m_Grid[5][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[6][nCol1] = m_Grid[6][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[7][nCol1] = m_Grid[7][nCol2];
#pragma warning(suppress: 26446 26482)
    m_Grid[8][nCol1] = m_Grid[8][nCol2];

#pragma warning(suppress: 26446 26482)
    m_Grid[0][nCol2] = tempCol[0];
#pragma warning(suppress: 26446 26482)
    m_Grid[1][nCol2] = tempCol[1];
#pragma warning(suppress: 26446 26482)
    m_Grid[2][nCol2] = tempCol[2];
#pragma warning(suppress: 26446 26482)
    m_Grid[3][nCol2] = tempCol[3];
#pragma warning(suppress: 26446 26482)
    m_Grid[4][nCol2] = tempCol[4];
#pragma warning(suppress: 26446 26482)
    m_Grid[5][nCol2] = tempCol[5];
#pragma warning(suppress: 26446 26482)
    m_Grid[6][nCol2] = tempCol[6];
#pragma warning(suppress: 26446 26482)
    m_Grid[7][nCol2] = tempCol[7];
#pragma warning(suppress: 26446 26482)
    m_Grid[8][nCol2] = tempCol[8];
  }

  void SwapBlockOfRows(int nBlockRow1, int nBlockRow2) noexcept
  {
    //Validate our parameters
    assert(nBlockRow1 >= 0);
    assert(nBlockRow1 < 3);
    assert(nBlockRow2 >= 0);
    assert(nBlockRow2 < 3);
    assert(nBlockRow1 != nBlockRow2);

    int nRow1{nBlockRow1 * 3};
    int nRow2{nBlockRow2 * 3};
    SwapRows(nRow1, nRow2);
    ++nRow1;
    ++nRow2;
    SwapRows(nRow1, nRow2);
    ++nRow1;
    ++nRow2;
    SwapRows(nRow1, nRow2);
  }

  void SwapBlockOfColumns(int nBlockCol1, int nBlockCol2) noexcept
  {
    //Validate our parameters
    assert(nBlockCol1 >= 0);
    assert(nBlockCol1 < 3);
    assert(nBlockCol2 >= 0);
    assert(nBlockCol2 < 3);
    assert(nBlockCol1 != nBlockCol2);

    int nCol1{nBlockCol1 * 3};
    int nCol2{nBlockCol2 * 3};
    SwapColumns(nCol1, nCol2);
    ++nCol1;
    ++nCol2;
    SwapColumns(nCol1, nCol2);
    ++nCol1;
    ++nCol2;
    SwapColumns(nCol1, nCol2);
  }

  void Transpose() noexcept //The columns become the rows and vice versa
  {
    //Make a copy of this grid
    const CSudokuGrid copy{*this};

    //Update our cell values from the copy
    for (unsigned char i{0}; i<9; i++) //i is row index index
    {
      for (int j{0}; j<9; j++) //j is column index index
        SetCell(i, j, copy.GetCell(j, i));
    }
  }

  void CreateRandomSolvableGrid(int nRounds = 10) noexcept
  {
    //Start off with the standard solved grid
    CreateStandardSolvedGrid();

    //Shuffle the grid a number by running a number of transformations on it
    for (int i{0}; i<nRounds; i++)
    {
      //Randomly select 2 rows in the same group to shuffle
      int nRow1{rand() % 3};
      int nRow2{0};
      if (nRow1 == 0)
        nRow2 = (rand() % 2) + 1;
      else if (nRow1  == 1)
        nRow2 = (rand() % 2) ? 0 : 2;
      else
        nRow2 = (rand() % 2);
      int nMultiplier{rand() % 3};
      nRow1 += (nMultiplier * 3);
      nRow2 += (nMultiplier * 3);
      SwapRows(nRow1, nRow2);

      //Do a transpose
      Transpose();

      //Randomly select 2 columns in the same group to shuffle
      int nCol1{rand() % 3};
      int nCol2{0};
      if (nCol1 == 0)
        nCol2 = (rand() % 2) + 1;
      else if (nCol1  == 1)
        nCol2 = (rand() % 2) ? 0 : 2;
      else
        nCol2 = (rand() % 2);
      nMultiplier = (rand() % 3);
      nCol1 += (nMultiplier * 3);
      nCol2 += (nMultiplier * 3);
      SwapColumns(nCol1, nCol2);

      //Do a transpose
      Transpose();

      //Randomly select 2 row blocks to shuffle
      nRow1 = rand() % 3;
      if (nRow1 == 0)
        nRow2 = (rand() % 2) + 1;
      else if (nRow1  == 1)
        nRow2 = (rand() % 2) ? 0 : 2;
      else
        nRow2 = (rand() % 2);
      SwapBlockOfRows(nRow1, nRow2);

      //Do a transpose
      Transpose();

      //Randomly select 2 column blocks to shuffle
      nCol1 = rand() % 3;
      if (nCol1 == 0)
        nCol2 = (rand() % 2) + 1;
      else if (nCol1  == 1)
        nCol2 = (rand() % 2) ? 0 : 2;
      else
        nCol2 = (rand() % 2);
      SwapBlockOfColumns(nCol1, nCol2);
    }
  }

  bool FillCellWithRandomLegalValue(int nRow, int nCol) noexcept
  {
    //What will be the return value from this method
    bool bUpdated{false};

    //Try a value from 1 to 9
    for (unsigned char k{1}; (k <= 9) && !bUpdated; k++)
    {
      if (IsLegalColumn(nRow, nCol, k) && IsLegalRow(nRow, nCol, k) && IsLegalBlock(nRow, nCol, k))
      {
        SetCell(nRow, nCol, k);
        bUpdated = true;
      }
    }

    return bUpdated;
  }

  int FillInAllDefiniteCells() noexcept
  {
    //What will be the return value from this method
    int nCellsUpdated{0};

    for (int i{0}; i<9; i++) //i is row index index
    {
      for (int j{0}; j<9; j++) //j is column index index
      {
        unsigned char byNumber{0};
        if (ChoicesForCell(i, j, &byNumber) == 1)
        {
          ++nCellsUpdated;
          SetCell(i, j, byNumber);
        }
      }
    }

    return nCellsUpdated;
  }

//Member variables
  std::array<std::array<unsigned char, 9>, 9> m_Grid{};
};
