#include "stdafx.h"
#include "SudokuSolver.h"
#include "SudokuSolverDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


constexpr const UINT WM_SOLVE_THREAD_FINISHED = WM_APP + 1;
constexpr const UINT WM_SOLVE_THREAD_STATUS_UPDATE = WM_APP + 2;


class CAboutDlg : public CDialog
{
public:
//Constructors / Destructors
  CAboutDlg();
  CAboutDlg(const CAboutDlg&) = delete;
  CAboutDlg(CAboutDlg&&) = delete;
  ~CAboutDlg() = default; //NOLINT(modernize-use-override)

//Methods
  CAboutDlg& operator=(const CAboutDlg&) = delete;
  CAboutDlg& operator=(CAboutDlg&&) = delete;

//Member variables
  enum { IDD = IDD_ABOUTBOX };

protected:
//Methods
  void DoDataExchange(CDataExchange* pDX) override;

  DECLARE_MESSAGE_MAP()
};


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CAboutDlg, CDialog) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

#pragma warning(suppress: 26455)
CAboutDlg::CAboutDlg() : CDialog{CAboutDlg::IDD}
{
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{

  __super::DoDataExchange(pDX);
}


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CSudokuSolverDlg, CDialog) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_SYSCOMMAND()
  ON_WM_PAINT()
  ON_WM_QUERYDRAGICON()
  ON_BN_CLICKED(IDC_HINT, &CSudokuSolverDlg::OnBnClickedHint)
  ON_BN_CLICKED(IDC_SOLVE, &CSudokuSolverDlg::OnBnClickedSolve)
  ON_BN_CLICKED(IDC_CLEAR, &CSudokuSolverDlg::OnBnClickedClear)
  ON_BN_CLICKED(IDC_GENERATE, &CSudokuSolverDlg::OnBnClickedGenerate)
  ON_MESSAGE(WM_SOLVE_THREAD_FINISHED, &CSudokuSolverDlg::OnThreadFinished)
  ON_MESSAGE(WM_SOLVE_THREAD_STATUS_UPDATE, &CSudokuSolverDlg::OnThreadStatusUpdate)
END_MESSAGE_MAP()

#pragma warning(suppress: 26455)
CSudokuSolverDlg::CSudokuSolverDlg(_In_opt_ CWnd* pParent) : CDialog{CSudokuSolverDlg::IDD, pParent},
                                                             m_pWorkerThread{nullptr},
                                                             m_StopSolveEvent{FALSE, TRUE},
                                                             m_hIcon{AfxGetApp()->LoadIcon(IDR_MAINFRAME)},
                                                             m_bSolvedGridValid{false},
                                                             m_nActiveCellIndex{-1},
                                                             m_nActiveDepth{0},
                                                             m_bHinting{false},
                                                             m_nLastStatusUpdate{0},
                                                             m_nEmptyCells{20},
                                                             m_SolvedCellOrder{}
{
}

CSudokuSolverDlg::~CSudokuSolverDlg()
{
#pragma warning(suppress: 26447)
  StopWorkerThread();
}

BOOL CSudokuSolverDlg::OnInitDialog()
{
  //Let the base class do its thing
  __super::OnInitDialog();

  //Add "About..." menu item to system menu.

  //IDM_ABOUTBOX must be in the system command range.
  ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
  ASSERT(IDM_ABOUTBOX < 0xF000);

  CMenu* pSysMenu{GetSystemMenu(FALSE)};
  if (pSysMenu != nullptr)
  {
    CString sAboutMenu;
    if (sAboutMenu.LoadString(IDS_ABOUTBOX) && !sAboutMenu.IsEmpty())
    {
      pSysMenu->AppendMenu(MF_SEPARATOR);
      pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, sAboutMenu);
    }
  }

  //Set the icon for this dialog.  The framework does this automatically
  //when the application's main window is not a dialog
  SetIcon(m_hIcon, TRUE); //Set big icon
  SetIcon(m_hIcon, FALSE); //Set small icon

  for (int i{0}; i<9; i++) //i is row index
  {
    for (int j{0}; j<9; j++) //j is column index
      SetDlgItemText(RowAndColumnToCellIdentifier(i, j), _T(""));
  }

  return TRUE;
}

#pragma warning(suppress: 26434)
void CSudokuSolverDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
  if ((nID & 0xFFF0) == IDM_ABOUTBOX)
  {
    CAboutDlg dlgAbout;
    dlgAbout.DoModal();
  }
  else
  {
    //Let the base class do its thing
    __super::OnSysCommand(nID, lParam);
  }
}

#pragma warning(suppress: 26434)
void CSudokuSolverDlg::OnPaint()
{
  if (IsIconic())
  {
    CPaintDC dc{this};

#pragma warning(suppress: 26490)
    SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

    //Center icon in client rectangle
    const int cxIcon{GetSystemMetrics(SM_CXICON)};
    const int cyIcon{GetSystemMetrics(SM_CYICON)};
    CRect rect;
    GetClientRect(&rect);
    const int x{(rect.Width() - cxIcon + 1) / 2};
    const int y{(rect.Height() - cyIcon + 1) / 2};

    //Draw the icon
    dc.DrawIcon(x, y, m_hIcon);
  }
  else
  {
    //Let the base class do its thing
    __super::OnPaint();
  }
}

#pragma warning(suppress: 26434 26440)
HCURSOR CSudokuSolverDlg::OnQueryDragIcon()
{
#pragma warning(suppress: 26473)
  return static_cast<HCURSOR>(m_hIcon);
}

void CSudokuSolverDlg::DoDataExchange(CDataExchange* pDX)
{
  //Let the base class do its thing
  __super::DoDataExchange(pDX);

  DDX_Control(pDX, IDC_SOLVE, m_wndSolve);
  DDX_Control(pDX, IDC_HINT, m_wndHint);
  DDX_Control(pDX, IDC_STATUS, m_wndStatus);
  DDX_Text(pDX, IDC_EMPTY_CELLS, m_nEmptyCells);
  DDV_MinMaxInt(pDX, m_nEmptyCells, 1, 81);

  //Handle all the cell values
  int nIndex{0};
  for (int i{0}; i<9; i++) //i is row index
  {
    for (int j{0}; j<9; j++) //J is the column index
    {
      int nCellIdentifier{RowAndColumnToCellIdentifier(i, j)};
#pragma warning(suppress: 26446 26482)
      DDX_Control(pDX, nCellIdentifier, m_wndEdits[nIndex]);

      if (pDX->m_bSaveAndValidate)
      {
        CString sText;
        GetDlgItemText(nCellIdentifier, sText);
        if (sText.IsEmpty())
          m_GridOnScreen.SetCellEmpty(i, j);
        else
        {
          int nValue{0};
          DDX_Text(pDX, nCellIdentifier, nValue);
          DDV_MinMaxInt(pDX, nValue, 1, 9);
#pragma warning(suppress: 26472)
          m_GridOnScreen.SetCell(i, j, static_cast<unsigned char>(nValue));
        }
      }
      else
      {
        int nValue{m_GridOnScreen.GetCell(i, j)};
        if (nValue == 0)
          SetDlgItemText(nCellIdentifier, _T(""));
        else
        {
          DDX_Text(pDX, nCellIdentifier, nValue);
          DDV_MinMaxInt(pDX, nValue, 1, 9);
        }
      }

      ++nCellIdentifier;
      ++nIndex;
    }
  }

  if (pDX->m_bSaveAndValidate)
    ValidateGridOnScreen(pDX);
}

#pragma warning(suppress: 26429)
void CSudokuSolverDlg::ValidateGridOnScreen(_In_ CDataExchange* pDX)
{
  //Step 1. Make sure all rows contain unique numbers
  for (int i{0}; i<9; i++) //i is row index
  {
    for (int j{0}; j<9; j++) //j is column index
    {
      if (!m_GridOnScreen.IsCellEmpty(i, j) && !m_GridOnScreen.IsLegalRow(i, j, m_GridOnScreen.GetCell(i, j)))
      {
#pragma warning(suppress: 26486)
        pDX->PrepareEditCtrl(RowAndColumnToCellIdentifier(i, j));
        AfxMessageBox(_T("Please enter a unique value in this cell for this row"));
        pDX->Fail();
      }
    }
  }

  //Step 2. Make sure all columns contain unique numbers
  for (int i{0}; i<9; i++) //i is row index
  {
    for (int j{0}; j<9; j++) //j is column index
    {
      for (int k{0}; k<9; k++) //j is row index index
      {
        if (!m_GridOnScreen.IsCellEmpty(i, j) && !m_GridOnScreen.IsLegalColumn(i, j, m_GridOnScreen.GetCell(i, j)))
        {
#pragma warning(suppress: 26486)
          pDX->PrepareEditCtrl(RowAndColumnToCellIdentifier(i, j));
          AfxMessageBox(_T("Please enter a unique value in this cell for this column"));
          pDX->Fail();
        }
      }
    }
  }

  //Step 3. Make sure all cells are unique in the 3 by 3 sudoku block
  for (int i{0}; i<9; i++) //i is row index
  {
    for (int j{0}; j<9; j++) //j is column index
    {
      if (!m_GridOnScreen.IsCellEmpty(i, j) && !m_GridOnScreen.IsLegalBlock(i, j, m_GridOnScreen.GetCell(i, j)))
      {
#pragma warning(suppress: 26486)
        pDX->PrepareEditCtrl(RowAndColumnToCellIdentifier(i, j));
        AfxMessageBox(_T("Please enter a unique value in this cell for this Sudoku block"));
        pDX->Fail();
      }
    }
  }
}

bool CSudokuSolverDlg::Solve(_In_ int nDepth, _Inout_ int& nActiveIteration, _In_ unsigned char nSolvedCells, _Inout_ CSudokuGrid& currentNode)
{
  //Step 1. Fill in all definite cells for the current grid. Also update the solved cell order matrix as we go
  for (int i{0}; i<9; i++) //i is row index index
  {
    for (int j{0}; j<9; j++) //j is column index index
    {
      unsigned char byNumber{0};
      if (currentNode.ChoicesForCell(i, j, &byNumber) == 1)
      {
        currentNode.SetCell(i, j, byNumber);
#pragma warning(suppress: 26446 26482)
        m_SolvedCellOrder[i][j] = nSolvedCells;
        ++nSolvedCells;
      }
    }
  }

  //Step 2. Check the current grid is solved
  if (currentNode.IsSolved())
    return true;

  //Step 3. Build up the children of the current grid in order of least choices
  bool bContinue{true};
  while (bContinue)
  {
    //Check to see if we have been requested to exit solving
    if (WaitForSingleObject(m_StopSolveEvent, 0) == WAIT_OBJECT_0)
      bContinue = false;
    else
    {
      //Find the cell to modify which has the least possible choices
      int nRow{-1};
      int nCol{-1};
      int nChoices{0};
      const bool bFound{currentNode.FindEmptyCellWithLeastChoice(nRow, nCol, nChoices)};
      if (!bFound || (nChoices == 0))
        bContinue = false;
      else
      {
        //Update the status UI every so often
        {
          CSingleLock cs{&m_csStatus, TRUE};
          m_sStatus.Format(_T("Solution depth:%d, Iteration:%d, Row:%d, Col:%d\n"), nDepth, nActiveIteration, nRow, nCol);
          m_nActiveDepth = nDepth;
          const ULONGLONG nCurrentTickCount{GetTickCount64()};
          if (GetSafeHwnd() && ((nCurrentTickCount - m_nLastStatusUpdate) > 100))
          {
            m_nLastStatusUpdate = nCurrentTickCount;
            PostMessage(WM_SOLVE_THREAD_STATUS_UPDATE, nRow, nCol);
          }
        }

        //Add a node to the tree for each of the possible moves for the specified cell
        for (unsigned char k{1}; (k<=9); k++)
        {
          if (currentNode.IsLegalColumn(nRow, nCol, k) && currentNode.IsLegalRow(nRow, nCol, k) && currentNode.IsLegalBlock(nRow, nCol, k))
          {
            ++nActiveIteration;

            //Step 4. Call ourselves recursively for each child we create

            //Make a copy of the current grid
            CSudokuGrid child{currentNode};

            //Modify the single cell
            child.SetCell(nRow, nCol, k);

            //Call ourselves recursively for this new child
            if (Solve(nDepth + 1, nActiveIteration, nSolvedCells, child))
            {
              currentNode = child;
              return true;
            }
          }
        }
      }
    }
  }

  return false;
}

UINT CSudokuSolverDlg::_WorkerThread(_In_opt_ LPVOID pParam)
{
  //Convert from the SDK world to the C++ world
#pragma warning(suppress: 26429)
  auto pThis{static_cast<CSudokuSolverDlg*>(pParam)};
  ASSERT(pThis != nullptr);

  return pThis->WorkerThread();
}

UINT CSudokuSolverDlg::WorkerThread()
{
  //Start to solve
  int nActiveIteration{0};
  const bool bSolved{Solve(1, nActiveIteration, 1, m_GridToSolve)};
  if (bSolved)
    m_SolvedGrid = m_GridToSolve;

  //Inform the UI thread if we have one
  if (GetSafeHwnd())
    PostMessage(WM_SOLVE_THREAD_FINISHED, bSolved);

  return 0;
}

LRESULT CSudokuSolverDlg::OnThreadFinished(WPARAM wParam, LPARAM /*lParam*/)
{
  if (m_pWorkerThread != nullptr)
  {
    //Kill the worker thread
    StopWorkerThread();

    //Reset the last active cell if there was one
    if (m_nActiveCellIndex != -1)
    {
      ASSERT((m_nActiveCellIndex >= 0) && (m_nActiveCellIndex < 81));
#pragma warning(suppress: 26446 26482)
      m_wndEdits[m_nActiveCellIndex].SetBKColor(GetSysColor(COLOR_WINDOW));
    }

    if (wParam) //wParam being non zero indicates we have solved the puzzle
    {
      m_bSolvedGridValid = true;

      if (m_bHinting)
        UpdateNextHintCellFromSolvedGrid();
      else
        UpdateAllCellsFromSolvedGrid();
    }
    else
    {
      m_wndStatus.SetWindowText(_T("Unable to solve the puzzle"));

      //Refresh the values on screen from their in memory values
      UpdateData(FALSE);
    }
  }

  return 0L;
}

LRESULT CSudokuSolverDlg::OnThreadStatusUpdate(WPARAM wParam, LPARAM lParam)
{
  //Reset the last active cell if there was one
  if (m_nActiveCellIndex != -1)
  {
    ASSERT(m_nActiveCellIndex >= 0 && m_nActiveCellIndex < 81);
#pragma warning(suppress: 26446 26482)
    m_wndEdits[m_nActiveCellIndex].SetBKColor(GetSysColor(COLOR_WINDOW));
#pragma warning(suppress: 26446 26482)
    m_wndEdits[m_nActiveCellIndex].SetWindowText(_T(""));
  }

  //Highlight the cell currently being operated on
#pragma warning(suppress: 26472)
  const int nCellIndex{(static_cast<int>(wParam) * 9) + static_cast<int>(lParam)};
  ASSERT(nCellIndex >= 0 && nCellIndex < 81);
#pragma warning(suppress: 26446 26482)
  m_wndEdits[nCellIndex].SetBKColor(RGB(255, 0, 0));
  m_nActiveCellIndex = nCellIndex;

  //Also update the status text and depth value
  CSingleLock cs{&m_csStatus, TRUE};
  m_wndStatus.SetWindowText(m_sStatus);
  CString sDepth;
  sDepth.Format(_T("[%u]"), m_nActiveDepth.operator ULONG());
#pragma warning(suppress: 26446 26482)
  m_wndEdits[nCellIndex].SetWindowText(sDepth);

  return 0L;
}

void CSudokuSolverDlg::StopWorkerThread()
{
  if (m_pWorkerThread != nullptr)
  {
    //Request the thread to exit
    m_StopSolveEvent.SetEvent();

    //Wait for the thread to fully exit
    WaitForSingleObject(m_pWorkerThread->m_hThread, INFINITE);
    delete m_pWorkerThread;
    m_pWorkerThread = nullptr;

    //Reset the status texts
    if (m_wndStatus.GetSafeHwnd())
      m_wndStatus.SetWindowText(_T("Stopped attempting to solve the puzzle"));
  }

  //Reset the button texts
  if (m_wndSolve.GetSafeHwnd())
    m_wndSolve.SetWindowText(_T("&Solve"));
  if (m_wndHint.GetSafeHwnd())
    m_wndHint.SetWindowText(_T("&Hint"));

  //Reset the colors for all the cells
  for (int i{0}; i<81; i++)
  {
#pragma warning(suppress: 26446 26482)
    if (m_wndEdits[i].GetSafeHwnd())
#pragma warning(suppress: 26446 26482)
      m_wndEdits[i].SetBKColor(GetSysColor(COLOR_WINDOW));
  }

  //Solved grid is now not valid
  m_bSolvedGridValid = false;
  m_SolvedCellOrder = {};
}

void CSudokuSolverDlg::StartWorkerThread(_In_ bool bHinting)
{
  //Setup the hinting boolean
  m_bHinting = bHinting;

  //Update the grid we are going to solve
  m_GridToSolve = m_GridOnScreen;

  //Reset the solved cell matrix
  m_bSolvedGridValid = false;
#pragma warning(suppress: 26485)
  m_SolvedCellOrder = {};

  //Spin of a worker thread for it (initially suspended so that we can setup it correctly!)
  ASSERT(m_pWorkerThread == nullptr);
  m_pWorkerThread = AfxBeginThread(_WorkerThread, this, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);
  if (m_pWorkerThread == nullptr)
    m_wndStatus.SetWindowText(_T("Failed to create worker thread"));
  else
  {
    //Update the button and status text
    if (bHinting)
      m_wndHint.SetWindowText(_T("&Cancel"));
    else
      m_wndSolve.SetWindowText(_T("&Cancel"));
    m_wndStatus.SetWindowText(_T("Solving puzzle, the cell being worked on will appear red and the depth of the algorithm will appear in that cell. Click the Cancel button at any time to abort"));

    //We're looking after its memory
    m_pWorkerThread->m_bAutoDelete = FALSE;

    //Reset the want stop variable
    m_StopSolveEvent.ResetEvent();

    //Everything ok, Resume the thread
    m_pWorkerThread->ResumeThread();
  }
}

void CSudokuSolverDlg::OnBnClickedSolve()
{
  if (m_pWorkerThread != nullptr)
  {
    //Kill the worker thread
    StopWorkerThread();

    //Refresh the values on screen from their in memory values
    UpdateData(FALSE);
  }
  else
  {
    if (UpdateData(TRUE))
    {
      //Check to see if the puzzle is already solved
      if (m_GridOnScreen.IsSolved())
      {
        //Update the status text
        m_wndStatus.SetWindowText(_T("Puzzle already solved!"));
        return;
      }

      //Fire off the worker thread to do the solving
      StartWorkerThread(false);
    }
  }
}

void CSudokuSolverDlg::OnBnClickedHint()
{
  if (m_pWorkerThread != nullptr)
  {
    //Kill the worker thread
    StopWorkerThread();

    //Refresh the values on screen from their in memory values
    UpdateData(FALSE);
  }
  else
  {
    if (UpdateData(TRUE))
    {
      //Check to see if the puzzle is already solved
      if (m_GridOnScreen.IsSolved())
      {
        //Update the status text
        m_wndStatus.SetWindowText(_T("Puzzle already solved!"));
        return;
      }

      //Look for the next empty cell which has only one possible choice
      int nRow{-1};
      int nCol{-1};
      int nChoices{0};
      unsigned char byNumber{0};
      if (m_GridOnScreen.FindEmptyCellWithLeastChoice(nRow, nCol, nChoices) && (nChoices == 1))
        m_GridOnScreen.ChoicesForCell(nRow, nCol, &byNumber);
      else
      {
        nRow = -1;
        nCol = -1;
      }

      //Fill in the next definite empty cell if we have one
      if (nRow != -1)
      {
        UpdateCell(nRow, nCol, byNumber, RGB(0, 255, 0));

        if (m_GridOnScreen.IsSolved())
        {
          //Update the status text
          m_wndStatus.SetWindowText(_T("Puzzle solved!"));
        }
        else
        {
          //Update the status text
          CString sStatus;
          sStatus.Format(_T("Updated Row:%d, Col:%d\n"), nRow, nCol);
          m_wndStatus.SetWindowText(sStatus);
        }
      }
      else if (m_bSolvedGridValid)
      {
        //Just update from the solved grid
        UpdateNextHintCellFromSolvedGrid();
      }
      else if (AfxMessageBox(_T("The puzzle needs to be solved to provide the next hint. Solve the puzzle now to provide the next hint?"), MB_YESNO) == IDYES)
      {
        //Fire off the worker thread to do the solving
        StartWorkerThread(true);
      }
    }
  }
}

void CSudokuSolverDlg::UpdateCell(_In_ int nRow, _In_ int nCol, _In_ unsigned char byNumber, _In_ COLORREF color)
{
  //Update the status text
  CString sStatus;
  sStatus.Format(_T("Updated Row:%d, Col:%d\n"), nRow, nCol);
  m_wndStatus.SetWindowText(sStatus);

  //Update the color
  const int nCellIndex{(nRow * 9) + nCol};
  ASSERT(nCellIndex >= 0 && nCellIndex < 81);
#pragma warning(suppress: 26446 26482)
  m_wndEdits[nCellIndex].SetBKColor(color);

  //Set the cell value in memory
  m_GridOnScreen.SetCell(nRow, nCol, byNumber);

  //Update the value value on screen
  CDataExchange DX{this, FALSE};
  int nValue{byNumber};
#pragma warning(suppress: 26486)
  DDX_Text(&DX, RowAndColumnToCellIdentifier(nRow, nCol), nValue);
}

void CSudokuSolverDlg::UpdateNextHintCellFromSolvedGrid()
{
  //Find the lowest value in m_SolvedCellOrder which corresponds to an empty cell
  int nRow{-1};
  int nCol{-1};
  int nLowestCellOrder{INT_MAX};
  for (int i{0}; i<9; i++) //i is row index index
  {
    for (int j{0}; j<9; j++) //j is column index index
    {
#pragma warning(suppress: 26446 26482)
      if (m_GridOnScreen.IsCellEmpty(i, j) && (m_SolvedCellOrder[i][j] < nLowestCellOrder))
      {
        nRow = i;
        nCol = j;
#pragma warning(suppress: 26446 26482)
        nLowestCellOrder = m_SolvedCellOrder[i][j];
      }
    }
  }

  //Update the next cell if found
  if (nRow != -1)
    UpdateCell(nRow, nCol, m_SolvedGrid.GetCell(nRow, nCol), RGB(0, 255, 0));
}

void CSudokuSolverDlg::UpdateAllCellsFromSolvedGrid()
{
  //Highlight all the updated cells in the solution in green
  for (int i{0}; i<9; i++) //i is row index
  {
    for (int j{0}; j<9; j++) //j is column index
    {
      if (m_GridOnScreen.IsCellEmpty(i, j))
        UpdateCell(i, j, m_SolvedGrid.GetCell(i, j), RGB(0, 255, 0));
    }
  }

  //Update the status text
  m_wndStatus.SetWindowText(_T("Puzzle solved!"));

  UpdateData(FALSE);
}

#pragma warning(suppress: 26440)
int CSudokuSolverDlg::RowAndColumnToCellIdentifier(_In_ int nRow, _In_ int nCol)
{
  //Validate our parameters
  ASSERT(nRow >= 0);
  ASSERT(nRow < 9);
  ASSERT(nCol >= 0);
  ASSERT(nCol < 9);

  return IDC_ROW1COL1 + (nRow * 9) + nCol;
}

void CSudokuSolverDlg::OnBnClickedClear()
{
  //Kill the worker thread if running
  StopWorkerThread();

  //Empty and reset the color for each cell
  for (int i{0}; i<81; i++)
  {
#pragma warning(suppress: 26446 26482)
    m_wndEdits[i].SetWindowText(_T(""));
#pragma warning(suppress: 26446 26482)
    m_wndEdits[i].SetBKColor(GetSysColor(COLOR_WINDOW));
  }

  m_wndStatus.SetWindowText(_T("Puzzle has been cleared"));
}

void CSudokuSolverDlg::OnBnClickedGenerate()
{
  OnBnClickedClear();

  //Collect the new amount of empty cells to create
  if (UpdateData(TRUE))
  {
    CWaitCursor wait;

    //Create a random solvable grid
    m_GridOnScreen.CreateRandomSolvableGrid();

    //Now empty out the specified number of cells
    int nCellsEmptied{0};
    while (nCellsEmptied < m_nEmptyCells)
    {
      const int i{rand() % 9};
      const int j{rand() % 9};

      if (!m_GridOnScreen.IsCellEmpty(i, j))
      {
        m_GridOnScreen.SetCellEmpty(i, j);
        ++nCellsEmptied;
      }
    }

    m_wndStatus.SetWindowText(_T("Puzzle has been generated"));

    UpdateData(FALSE);
  }
}



#pragma warning(suppress: 26433 26440 26477)
IMPLEMENT_DYNAMIC(CColorEdit, CEdit)

#pragma warning(suppress: 26455)
CColorEdit::CColorEdit() : m_clrBackground{GetSysColor(COLOR_WINDOW)}
{
  m_brBkgnd.CreateSolidBrush(m_clrBackground);
}

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CColorEdit, CEdit) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_CTLCOLOR_REFLECT()
END_MESSAGE_MAP()

#pragma warning(suppress: 26429)
HBRUSH CColorEdit::CtlColor(CDC* pDC, UINT /*nCtlColor*/)
{
  pDC->SetBkColor(m_clrBackground);
  return static_cast<HBRUSH>(m_brBkgnd.GetSafeHandle());
}

void CColorEdit::SetBKColor(_In_ COLORREF clrBackground)
{
  m_clrBackground = clrBackground;
  m_brBkgnd.DeleteObject();
  m_brBkgnd.CreateSolidBrush(m_clrBackground);
  Invalidate();
}
