#pragma once

#include "Sudoku.h"
#include <array>


class CColorEdit : public CEdit
{
public:
//Constructors / Destructors
  CColorEdit();
  CColorEdit(const CColorEdit&) = delete;
  CColorEdit(CColorEdit&&) = delete;
  ~CColorEdit() = default; //NOLINT(modernize-use-override)

//Methods
  CColorEdit& operator=(const CColorEdit&) = delete;
  CColorEdit& operator=(CColorEdit&&) = delete;
  void SetBKColor(_In_ COLORREF clrBackground);

protected:
//Member variables
  COLORREF m_clrBackground;
  CBrush m_brBkgnd;

//Message handlers
  afx_msg HBRUSH CtlColor(CDC* pDC, UINT nCtlColor);

  DECLARE_DYNAMIC(CColorEdit)

  DECLARE_MESSAGE_MAP()
};


class CSudokuSolverDlg : public CDialog
{
public:
//Constructors / Destructors
  CSudokuSolverDlg(_In_opt_ CWnd* pParent = nullptr);
  CSudokuSolverDlg(const CSudokuSolverDlg&) = delete;
  CSudokuSolverDlg(CSudokuSolverDlg&&) = delete;
  ~CSudokuSolverDlg(); //NOLINT(modernize-use-override)

//Methods
  CSudokuSolverDlg& operator=(const CSudokuSolverDlg&) = delete;
  CSudokuSolverDlg& operator=(CSudokuSolverDlg&&) = delete;

//Member variables
  enum { IDD = IDD_SUDOKUSOLVER_DIALOG };
  std::array<CColorEdit, 81> m_wndEdits;
  CButton m_wndSolve;
  CButton m_wndHint;
  CEdit m_wndStatus;

protected:
//Methods
  void DoDataExchange(CDataExchange* pDX) override;
  int RowAndColumnToCellIdentifier(_In_ int nRow, _In_ int nCol);
  BOOL OnInitDialog() override;
  void ValidateGridOnScreen(_In_ CDataExchange* pDX);
  void StartWorkerThread(_In_ bool bHinting);
  void StopWorkerThread();
  bool Solve(_In_ int nDepth, _Inout_ int& nActiveIteration, _In_ unsigned char nSolvedCells, _Inout_ CSudokuGrid& currentNode);
  static UINT _WorkerThread(_In_opt_ LPVOID pParam);
  UINT WorkerThread();
  void UpdateNextHintCellFromSolvedGrid();
  void UpdateAllCellsFromSolvedGrid();
  void UpdateCell(_In_ int nRow, _In_ int nCol, _In_ unsigned char byNumber, _In_ COLORREF color);

//Message handlers
  afx_msg void OnSysCommand(UINT nID, LPARAM lParam);
  afx_msg void OnPaint();
  afx_msg HCURSOR OnQueryDragIcon();
  afx_msg void OnBnClickedClear();
  afx_msg void OnBnClickedHint();
  afx_msg void OnBnClickedSolve();
  afx_msg void OnBnClickedGenerate();
  afx_msg LRESULT OnThreadFinished(WPARAM wParam, LPARAM lParam);
  afx_msg LRESULT OnThreadStatusUpdate(WPARAM wParam, LPARAM lParam);
  
  DECLARE_MESSAGE_MAP()
  
//Member variables
  CWinThread* m_pWorkerThread;
  CEvent m_StopSolveEvent;
  HICON m_hIcon;
  CSudokuGrid m_GridOnScreen;
  CSudokuGrid m_SolvedGrid;
  bool m_bSolvedGridValid;
  CCriticalSection m_csStatus;
  CString m_sStatus;
  int m_nActiveCellIndex;
  std::atomic<ULONG> m_nActiveDepth;
  bool m_bHinting;
  ULONGLONG m_nLastStatusUpdate;
  int m_nEmptyCells;
  std::array<std::array<unsigned char, 9>, 9> m_SolvedCellOrder;
  CSudokuGrid m_GridToSolve;
};
